#!/usr/bin/python
#-----------------------------------------------------------------------
# File    : detect.py
# Contents: detect synchronous spiking activity
# Author  : Christian Borgelt
# History : 2014.06.30 file created from cocoexp/detect.py
#           2014.07.09 result collection over signatures added
#           2014.07.10 several small bugs fixed
#           2015.01.16 adapted to new sequence construction functions
#           2015.01.22 overlap and additional patterns corrected
#-----------------------------------------------------------------------
from sys      import argv, path
from os       import makedirs
from os.path  import join
from math     import sqrt
from time     import strftime
from random   import seed
UTILDIR  = join('..', 'util')
if UTILDIR  not in path: path.insert(0, UTILDIR)
from id2par   import get_args, project, run
NEURODIR = join('..', 'neuro')
if NEURODIR not in path: path.insert(0, NEURODIR)
from psts     import genpsts
from hypernad import itemseq
try:
    from coco    import coconad, estpsp
except ImportError:
    COCODIR = join('..', 'coconad')
    if COCODIR not in path: path.insert(0, COCODIR)
    from coconad import coconad
    from estpsp  import estpsp
    print('Attention: Could not import from PyCoCo library!')
    print('           Using Python replacement (very slow)!')

#-----------------------------------------------------------------------

oo = float('inf')               # positive infinity as a constant

#-----------------------------------------------------------------------
# Detection Function
#-----------------------------------------------------------------------

def detect (ncnt=100, period=3, rate=20,          # neuron    parameters
            size=5, coin=5, msel=0, jitter=0.003, # assembly  parameters
            width=0.003, logth=0,                 # detection parameters
            modes=[('p',0)],                      # detection parameters
            runs=1000, verbose=10, srng=1):       # other     parameters
    if jitter < 0: jitter = width
    srng += 1023*size           # init. seed and get data file name
    name = 'n%d_p%g_r%g_m%d_j%g_w%g_t%g_%d_%d' \
         % (ncnt,period,rate,msel,jitter*1000,width*1000,logth,
            size,coin)          # create the file name
    asm  = set(range(1,size+1)) # create the assembly to inject
    thpsp  = 10**-logth         # get threshold for pattern spectrum
    result = dict([(m,[]) for m in modes])
    for k in range(runs):       # repeat the experiment multiple times
        if verbose > 0 and k % verbose == 0:
            print(strftime('%Y.%m.%d %H:%M:%S detect ')+name+' '+str(k))
        seed(srng)              # seed random number generator
        psts = genpsts(ncnt, period, rate, jitter=jitter,
                       asms=[[asm,-coin,1,(msel-coin)/float(coin)]])
        psts = list(zip(range(1,ncnt+1), psts))
        psp  = estpsp(psts, 's', width, 2, 2, -1, '#',
                      10000, 0.5, 1000, srng)
        psp  = dict([(s,c) for s,c in psp.items() if c >= thpsp])
        border = list(range(ncnt+1)); m = -1
        for z in reversed(border):
            m = max([c for a,c in psp if a == z]+[m])
            border[z] = m+1     # determine the decision border
        border[0:2] = [oo,oo]   # entirely rule out sizes 0 and 1
        pats = coconad(psts, 'c', width, 2, 2,
                       report='a', border=border)
        for m,e in modes:       # traverse the item sequence modes
            iseq = itemseq(m, e, pats, psts, width)
            best = max([(iseq[i][1],i) for i in range(len(iseq))]
                       +[(0,0)])[1]+1
            dect = set([i for i,c in iseq[:best]])
            nmis = len(asm-dect)# get the best item set at start
            nadd = len(dect-asm)# and check missing and add. items
            val  = iseq[best-1][1] if best <= len(iseq) else 0.0
            result[(m,e)].append((best, nmis, nadd, val))
        srng += coin            # switch to the next random seed
    if verbose > 0:             # print termination log message
        print(strftime('%Y.%m.%d %H:%M:%S detect ')
                      +name+' '+str(runs)+' done.')
    try: makedirs('data')       # ensure that a directory exists
    except OSError: pass        # for the result files
    n = 1.0/runs if runs > 0 else 1.0
    for m in modes:             # traverse the item sequence modes
        name = 'n%d_p%g_r%g_m%d_j%g_w%g_t%g_%c%d_%d_%d' \
               % (ncnt,period,rate,msel,jitter*1000,width*1000,logth,
                  m[0],m[1],size,coin)# create the file name
        with open(join('data', 'dc_'+name+'.txt'), 'w') as out:
            # exact   : none missing, no additional
            exa = n*len([b for b,mis,add,v in result[m]
                         if mis <= 0 and add <= 0])
            # superset: none missing, at least one additional
            sup = n*len([b for b,mis,add,v in result[m]
                         if mis <= 0 and add >  0])
            # subset  : at least one missing, at least one present,
            #           no additional
            sub = n*len([b for b,mis,add,v in result[m]
                         if mis >  0 and mis < size and add <= 0])
            # overlap : at least one missing, at least one present,
            #           at least one additional
            ovl = n*len([b for b,mis,add,v in result[m]
                         if mis > 0  and mis < size and add >  0])
            # unrel.  : at most one present, at least one additional
            url = n*len([b for b,mis,add,v in result[m]
                         if mis >= size-1 and add > 0])
            # empty   : all missing, no additional
            emt = n*len([b for b,mis,add,v in result[m]
                         if mis >= size and add <= 0])
            out.write('%g %g %g %g %g %g\n' % (exa,sup,sub,ovl,url,emt))
            for r in result[m]: out.write('%d %d %d %g\n' % r)
    return result               # return the result list

#-----------------------------------------------------------------------
# Collection Function
#-----------------------------------------------------------------------

def collect (ncnt=100, period=3, rate=20,         # neuron    parameters
             sizes=[5], coins=[5], msel=0,        # assembly  parameters
             jitter=0.003,                        # assembly  parameters
             width=0.003, logth=1,                # detection parameters
             modes=[('p',0)]):                    # detection parameters
    if jitter < 0: jitter = width
    for m,e in modes:           # traverse the item sequence modes
        base = 'n%d_p%g_r%g_m%d_j%g_w%g_t%g_%c%d' \
               % (ncnt,period,rate,msel,jitter*1000,width*1000,
                  logth,m,e)
        fneg = dict(); fnxs = dict()# initialize the dictionaries
        sups = dict(); subs = dict(); ovls = dict();
        urls = dict(); emts = dict()
        for s in [(z,c) for z in sizes for c in coins]:
            inpfn = join('data','dc_'+base+('_%d_%d.txt' % s))
            with open(inpfn, 'r') as inp:
                v = [float(x) for x in inp.read()[:-1].split()[:6]]
                fneg[s],fnxs[s] = 1.0-v[0]-v[1],1.0-v[0]
                sups[s],subs[s],ovls[s],urls[s],emts[s] = v[1:]
        with open(join('data','fn_' +base +'.txt'), 'w') as out:
            for x in sorted([(k[0], k[1], fneg[k]) for k in fneg]):
                out.write('%g %g %s\n' % x)
        with open(join('data','fx_' +base +'.txt'), 'w') as out:
            for x in sorted([(k[0], k[1], fnxs[k]) for k in fnxs]):
                out.write('%g %g %s\n' % x)
        with open(join('data','sp_' +base +'.txt'), 'w') as out:
            for x in sorted([(k[0], k[1], sups[k]) for k in sups]):
                out.write('%g %g %s\n' % x)
        with open(join('data','sb_' +base +'.txt'), 'w') as out:
            for x in sorted([(k[0], k[1], subs[k]) for k in subs]):
                out.write('%g %g %s\n' % x)
        with open(join('data','ov_' +base +'.txt'), 'w') as out:
            for x in sorted([(k[0], k[1], ovls[k]) for k in ovls]):
                out.write('%g %g %s\n' % x)
        with open(join('data','ur_' +base +'.txt'), 'w') as out:
            for x in sorted([(k[0], k[1], urls[k]) for k in urls]):
                out.write('%g %g %s\n' % x)
        with open(join('data','mt_' +base +'.txt'), 'w') as out:
            for x in sorted([(k[0], k[1], emts[k]) for k in emts]):
                out.write('%g %g %s\n' % x)

#-----------------------------------------------------------------------

if __name__ == '__main__':
    g     = [ ( 'ncnt',   [   100 ] ),
              ( 'period', [     3 ] ),
              ( 'rate',   [    20 ] ),
              ( 'jitter', [ 0.003 ] ),
              ( 'width',  [ 0.003 ] ),
              ( 'logth',  [ 0, 1, 2, 3 ] ),
              ( 'modes',  [ [('p',0),('p',1),
                             ('i',0),('i',1),
                             ('s',0),('s',1)] ] ),
              ( 'runs',   [  1000 ] ) ]
    grids = [ [ ( 'size', [ 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2 ] ),
                ( 'coin', [ 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2 ] ),
                ( 'msel', [ 0 ] ) ],
              [ ( 'size', [ 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2 ] ),
                ( 'coin', [ 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3 ] ),
                ( 'msel', [ 1 ] ) ],
              [ ( 'size', [ 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2 ] ),
                ( 'coin', [ 15, 14, 13,
                            12, 11, 10, 9, 8, 7, 6, 5, 4 ] ),
                ( 'msel', [ 2 ] ) ],
              [ ( 'size', [ 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2 ] ),
                ( 'coin', [ 17, 16, 15, 14, 13,
                            12, 11, 10, 9, 8, 7, 6, 5 ] ),
                ( 'msel', [ 3 ] ) ],
              [ ( 'size', [ 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2 ] ),
                ( 'coin', [ 19, 18, 17, 16, 15, 14, 13,
                            12, 11, 10, 9, 8, 7, 6 ] ),
                ( 'msel', [ 4 ] ) ],
              [ ( 'size', [ 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2 ] ),
                ( 'coin', [ 21, 20, 19, 18, 17, 16, 15, 14, 13,
                            12, 11, 10, 9, 8, 7 ] ),
                ( 'msel', [ 5 ] ) ] ]
    grids = [g+x for x in grids]

    if len(argv) < 2 or argv[1] != 'collect':
        run(detect, grids, int(argv[1])-1 if len(argv) > 1 else -1)
    else:
        for g in grids:
            ranges = dict(project([g], ['size','coin'])[0])
            g = project([g], ['ncnt','period','rate','msel','jitter',
                              'width','logth','modes'])
            for i in range(get_args(g)):
                collect(sizes=ranges['size'], coins=ranges['coin'],
                        **get_args(g, i))
