#!/usr/bin/python
#-----------------------------------------------------------------------
# File    : hypernad.py
# Contents: hypergraph-based neuron assembly detection with CoCoNAD
# Author  : Christian Borgelt
# History : 2014.05.30 file created
#           2014.06.04 item sequencing by backward elimination added
#           2014.06.05 pattern spectrum estimation and filtering added
#           2014.06.18 item evaluation functions added for selection
#           2014.06.25 only simplified selection with item instances
#           2015.01.19 new versions of event based methods added
#           2015.01.29 simplified to needed/well working functions
#-----------------------------------------------------------------------
from sys     import argv, stderr, path, version_info
from os.path import join
from time    import time
from math    import floor, ceil, sqrt, log
COCODIR = join('..', 'coconad')
if COCODIR not in path: path.insert(0, COCODIR)
from events  import events
try:
    from coco    import coconad, estpsp
except ImportError:
    from coconad import coconad
    from estpsp  import estpsp
    print('Attention: Could not import from PyCoCo library!')
    print('           Using Python replacement (very slow)!')

#-----------------------------------------------------------------------
# Constants
#-----------------------------------------------------------------------
oo = float('inf')               # positive infinity as a constant

#-----------------------------------------------------------------------
# File Reading Function
#-----------------------------------------------------------------------

def getrec (f, recseps, fldseps, blanks, comment):
    '''Read a record from a file and split it into fields.
f       file to read from
recseps record  separators
fldseps field   separators
blanks  blank   characters
comment comment characters
returns a list of the fields read or None at end of file'''
    c = f.read(1).decode()      # read the next character
    while c != '' and c in comment:
        while c != '' and c not in recseps:
            c = f.read(1).decode() # skip comment until record separator
        c = f.read(1).decode()  # consume the record separator
    if c == '': return None     # if at end of file, abort
    rec = []                    # initialize the record
    sep = fldseps +recseps      # combine all separators
    while 1:                    # field read loop
        while c != '' and c in blanks:
            c = f.read(1).decode() # skip leading blanks
        fld = ''                # initialize the field
        while c != '' and c not in sep:
            fld += c            # append character to the field
            c = f.read(1).decode()  # and read the next character
        fld = fld.strip(blanks) # remove leading/trailing blanks */
        if len(fld) > 0:        # if the field is not empty,
            rec.append(fld)     # append the field to the record
        if c == '' or c in recseps: break
        c = f.read(1).decode()  # consume field separator
    return rec                  # return the record

#-----------------------------------------------------------------------
# Event Evaluation Functions
#-----------------------------------------------------------------------

def eev_allcnt (events, item, n):
    return len(set([x for e in events for x in e]))

#-----------------------------------------------------------------------

def eev_redcnt (events, item, n):
    return len(set([x for e in events for x in e if x[0] != item]))

#-----------------------------------------------------------------------

eevfns = [ eev_allcnt, eev_redcnt ]

#-----------------------------------------------------------------------
# Item Sequence Functions
#-----------------------------------------------------------------------

def seqpats (pats, parm=1):
    '''Build a item sequence by backward elimination.
pats    frequent patterns as a list of pairs (I,s), where I is a list
        of items (the item set) and s the support of the item set
parm    parameter (number of items to be seen as a reference)
returns a list of items with as pairs of an item and a list of values
        associated with the item.'''
    # print('seqpats: %d' % parm)
    items = set([i for p,c in pats for i in p])
    pats  = [(frozenset(p),c*(len(p)-parm)) for p,c in pats]
    iseq  = []                  # turn item sets into actual sets
    while 1:                    # while to remove more items
        pats = [p for p in pats if p[0] <= items]
        if not pats: break      # reduce patterns, init. evaluations
        vals = dict([(i,0) for i in items])
        for p,v in pats:        # sum item instance counts per item
            for i in p: vals[i] += v
        n = min(vals, key=lambda x: vals[x])
        v = vals[n]             # find the least connected item and
        iseq.append((n,v))      # add the item to the sequence, then
        items.remove(n)         # remove it from the remaining items
    iseq += [(i,0) for i in items]
    iseq.reverse()              # add last items and reverse seq.
    return iseq                 # return the reversed sequence

#-----------------------------------------------------------------------

def seqiovl (pats, trains, width=0.003, parm=1):
    '''Build a item sequence by backward elimination.
pats    frequent patterns as a list of pairs (I,s), where I is a list
        of items (the item set) and s the support of the item set
trains  underlying events trains in which the patterns were found
width   window width with which the patterns were found
parm    parameter of event set evaluation function
returns a list of items with as pairs of an item and a list of values
        associated with the item.'''
    # print('seqiovl: %d' % parm)
    eevfn  = eevfns[parm]       # get the evaluation function
    items  = set([i for p,c in pats for i in p])
    trains = dict(trains)       # collect items in all patterns
    pats   = [(frozenset(p),events([(i,trains[i]) for i in p],width,0))
              for p,c in pats]  # collect events underlying item sets
    iseq   = []                 # initialize the item sequence
    while 1:                    # while to remove more items
        pats = [p for p in pats if p[0] <= items]
        if not pats: break      # reduce patterns, init. evaluations
        cnts = dict()           # init. the event counters
        for p,e in pats:        # traverse the patterns
            for g in e:         # traverse the event groups
                for x in g:     # traverse and count the event occs.
                    if x in cnts: cnts[x] += 1
                    else:         cnts[x]  = 1
        evts = [(p,[g for g in e if [x for x in g if cnts[x] > 1]])
                for p,e in pats]# reduce to event groups with overlap
        vals = dict([(n,[]) for n in items])
        for p,e in evts:        # collect connecting events per item
            for i in p: vals[i].extend(e)
        for i in vals:          # get number of connecting events
            vals[i] = eevfn(vals[i], i, len(items))
        n = min(vals, key=lambda x: vals[x])
        c = vals[n]             # find the least connected item and
        iseq.append((n,c))      # add the item to the sequence, then
        items.remove(n)         # remove it from the connected comp.
    iseq += [(i,0) for i in items]
    iseq.reverse()              # add last two items and reverse seq.
    return iseq                 # return the reversed sequence

#-----------------------------------------------------------------------

def seqisyn (pats, trains, width=0.003, parm=1):
    '''Build a item sequence by backward elimination.
pats    frequent patterns as a list of pairs (I,s), where I is a list
        of items (the item set) and s the support of the item set
trains  underlying events trains in which the patterns were found
width   window width with which the patterns were found
parm    identifier of event set evaluation function
returns a list of items with as pairs of an item and a list of values
        associated with the item.'''
    # print('seqisyn: %d' % parm)
    eevfn  = eevfns[parm]       # get the evaluation function
    items  = set([i for p,c in pats for i in p])
    trains = dict(trains)       # collect items in all patterns
    pats   = [(frozenset(p),events([(i,trains[i]) for i in p],width,0))
              for p,c in pats]  # collect events underlying item sets
    pats   = [(p,[[sorted(g, key=lambda x: x[1]),0] for g in e])
              for p,e in pats]  # sort the events groups by time
    iseq   = []                 # initialize the item sequence
    while 1:                    # while to remove more items
        pats = [p for p in pats if p[0] <= items]
        if not pats: break      # reduce patterns, init. evaluations
        for p,e in pats:        # clear the overlap indicators
            for g in e: g[1] = 0
        grps = [g for p,e in pats for g in e]
        grps.sort(key=lambda y:(y[0][0],y[0][-1]))
        n = len(grps)           # sort groups by first and last event
        for i in range(n):      # traverse the event groups
            g = grps[i]         # get the next group
            if g[1]: continue   # if already marked, skip group
            if i > 0:           # if not the first group
                h = grps[i-1]   # check against previous group
                if g[0][-1][1] -h[0][0][1] < width:
                    g[1] = h[1] = 1
            if i < n-1:         # if not the last group
                h = grps[i+1]   # check against following group
                if h[0][-1][1] -g[0][0][1] < width:
                    g[1] = h[1] = 1
        evts = [(p,[g[0] for g in e if g[1]]) for p,e in pats]
        vals = dict([(n,[]) for n in items])
        for p,e in evts:        # collect connecting events per item
            for i in p: vals[i].extend(e)
        for i in vals:          # get number of connecting events
            vals[i] = eevfn(vals[i], i, len(items))
        n = min(vals, key=lambda x: vals[x])
        c = vals[n]             # find the least connected item and
        iseq.append((n,c))      # add the item to the sequence, then
        items.remove(n)         # remove it from the connected comp.
    iseq += [(i,0) for i in items]
    iseq.reverse()              # add last two items and reverse seq.
    return iseq                 # return the reversed sequence

#-----------------------------------------------------------------------

def itemseq (mode, parm, pats, trains, width=0.003):
    if mode == 'p': return seqpats(pats, parm)
    if mode == 'i': return seqiovl(pats, trains, width, parm)
    if mode == 's': return seqisyn(pats, trains, width, parm)

#-----------------------------------------------------------------------
# Main Program
#-----------------------------------------------------------------------

def error (msg):
    stderr.write(msg); exit()   # print an error message and abort

#-----------------------------------------------------------------------

if __name__ == '__main__':
    desc    = 'hypergraph-based neuron assembly detection with CoCoNAD'
    version = 'version 2.0 (2015.01.29)         ' \
            + '(c) 2014   Christian Borgelt'
    args    = []                  # list of program options
    opts    = {'t': ['target', 'c' ],  # target type
               'w': ['width', 0.003],  # width of window/region
               's': ['supp',  2.0  ],  # minimum support (absolute)
               'm': ['zmin',  2    ],  # minimum number of items
               'n': ['zmax', -1    ],  # maximum number of items
               'a': ['beg',  -oo   ],  # minimum of point/time range
               'z': ['end',  +oo   ],  # minimum of point/time range
               'e': ['alpha', 0.5  ],  # probability dispersion factor
               'i': ['smpls', 1000 ],  # samples per item set size
               'c': ['thpsp', 10.0 ],  # threshold for pattern spectrum
               'S': ['seed',  0    ],  # seed for random numbers
               'q': ['mode',  'p'  ],  # item sequence building mode
               'p': ['parm',  0    ],  # evaluation function parameter
               'l': ['train', False],  # one train per line/record
               'y': ['item',  True ],  # item in line, item first
               'h': ['ishdr', ''   ],  # record header  for output
               'k': ['isep',  ' '  ],  # item separator for output
               'v': ['outfmt',' (%g)'],# output format for support
               'r': ['recseps', '\\n' ],     # record  separators
               'f': ['fldseps', ' ,\\t' ],   # field   separators
               'b': ['blanks',  ' \\t\\r' ], # blank   characters
               'C': ['comment', '#' ] }      # comment characters

    if len(argv) <= 1:         # if no arguments are given
        opts = dict([(o,opts[o][1]) for o in opts])
        print('usage: hypernad.py [options] infile [outfile [seqfile]]')
        print(desc); print(version)
        print('-t#      target type for pattern mining         '
                      +'(default: %s)' % opts['t'])
        print('-w#      width of time window/maximum distance  '
                      +'(default: %g)' % opts['w'])
        print('-s#      minimum support of an item set         '
                      +'(default: %g)' % opts['s'])
        print('-m#      minimum number of items per item set   '
                      +'(default: %d)' % opts['m'])
        print('-n#      maximum number of items per item set   '
                      +'(default: no limit)')
        print('-a#      start of point/time range              '
                      +'(default: auto)')
        print('-z#      end   of point/time range              '
                      +'(default: auto)')
        print('-e#      probability dispersion factor          '
                      +'(default: %g)' % opts['e'])
        print('-i#      number of item set samples per size    '
                      +'(default: %d)' % opts['i'])
        print('-c#      threshold for pattern spectrum         '
                      +'(default: %d)' % opts['c']);
        print('-S#      seed for pseudo-random numbers         '
                      +'(default: %d)' % opts['S'])
        print('-q#      item sequence building mode            '
                      +'(default: %s)' % opts['q'])
        print('-p#      evaluation function parameter          '
                      +'(default: %d)' % opts['p'])
        print('-l       one train per record (item and points) '
                      +'(default: id/time pairs)')
        print('-y       no item (with -l) or item after point  '
                      +'(default: item first)')
        print('-r#      record  separators                     '
                      +'(default: "%s")' % opts['r'])
        print('-f#      field   separators                     '
                      +'(default: "%s")' % opts['f'])
        print('-b#      blank   characters                     '
                      +'(default: "%s")' % opts['b'])
        print('-C#      comment characters                     '
                      +'(default: "%s")' % opts['C'])
        print('infile   file to read (spike) trains from       '
                      +'[required]')
        print('outfile  file to write found patterns to        '
                      +'[optional]')
        print('seqfile  file to write found item sequence to   '
                      +'[optional]')
        exit()                  # print usage message and abort

    stderr.write('hypernad.py') # print a startup message
    stderr.write(' - ' +desc +'\n' +version +'\n')

    # --- evaluate program options and arguments ---
    for a in argv[1:]:          # traverse the program arguments
        if a[0] != '-':         # check for a fixed argument
            if len(args) < 3: args.append(a); continue
            else: error('too many arguments\n')
        if a[1] not in opts:    # check for an option
            error('unknown option: %s\n' % a[:2])
        o = opts[a[1]]          # get the corresponding option
        v = a[2:]               # and get the option argument
        if   isinstance(o[1],True.__class__): o[1] = not o[1]
        elif isinstance(o[1], (0).__class__): o[1] = int(v)
        elif isinstance(o[1], 0.0.__class__): o[1] = float(v)
        else:                                 o[1] = v
    if len(args) < 1: error('missing arguments\n')
    opts = dict([opts[o] for o in opts])

    target  = opts['target']    # target type for CoCoNAD
    zmin    = opts['zmin']      # minimum size of item sets
    zmax    = opts['zmax']      # maximum size of item sets
    supp    = opts['supp']      # minimum support for CoCoNAD
    width   = opts['width']     # time window width/maximum distance
    beg     = opts['beg']       # beginning of allowed span of spikes
    end     = opts['end']       # end       of allowed span of spikes
    alpha   = opts['alpha']     # probability dispersion factor
    smpls   = opts['smpls']     # number of item set samples per size
    thpsp   = opts['thpsp']     # threshold for pattern spectrum
    seed    = opts['seed']      # random seed (0: use time)
    mode    = opts['mode']      # item sequence building mode
    parm    = opts['parm']      # evaluation function parameter
    tpr     = opts['train']     # flag for one train per line
    item    = opts['item']      # and flag for item first
    ishdr   = opts['ishdr']     # record header  for output
    isep    = opts['isep']      # item separator for output
    outfmt  = opts['outfmt']    # output format for support
    recseps = opts['recseps']   # record  separators
    fldseps = opts['fldseps']   # field   separators
    blanks  = opts['blanks']    # blank   characters
    comment = opts['comment']   # comment characters
    if supp  <  0: supp = -supp  # check the parameters
    if zmin  <= 0: error('invalid minimum size %d\n' % zmin)
    if width <  0: error('invalid window width %g\n' % width)
    if parm  <  0 or parm > 1:
        error('invalid evaluation function parameter %d\n' % parm)
    x = [recseps,fldseps,blanks,comment]
    if version_info[0] >= 3:    # decode ASCII escape sequences
        x = [bytes(s, 'utf-8').decode('unicode_escape') for s in x]
    else:                       # decode ASCII escape sequences
        x = [s.decode('string_escape') for  s in x]
    recseps,fldseps,blanks,comment = x

    # --- read the data set to analyze ---
    t = time()                  # start timer, print log message
    stderr.write('reading %s ... ' % args[0])
    trains = dict(); k = 1      # initialize the train dictionary
    with open(args[0], 'rb') as inp:
        while 1:                # record read loop
            rec = getrec(inp, recseps, fldseps, blanks, comment)
            if rec == None: break
            if not rec: continue# read the next record
            if tpr:             # if one train per line/record
                if item: n = rec[0]; rec = rec[1:]
                else:    n = k+1# get the item/neuron identifier
                trains[n] = [x for x in [float(x) for x in line]
                             if x >= beg and x <= end]
                k += 1          # collect the trains
            else:               # if item/ point pairs
                if item: n,s = rec[:2]
                else:    s,n = rec[:2]
                s = float(s)    # collect points per item/neuron
                if s < beg or s > end: continue
                if n in trains: trains[n].append(s)
                else:           trains[n] = [s]
    trains = trains.items()     # convert to pairs (item, list of times)
    points = sum([len(s) for n,s in trains])
    stderr.write('[%d item(s), %d point(s)]' % (len(trains), points))
    stderr.write(' done [%.2fs].\n' % (time()-t))
    if beg <= -oo: beg = floor(min(min(t) for n,t in trains))
    if end >= +oo: end = ceil (max(max(t) for n,t in trains))

    # --- read or generate pattern spectrum ---
    t = time()                  # start timer, print log message
    stderr.write('estimating pattern spectrum ... '); stderr.flush()
    psp = estpsp(trains, 's', width, supp, zmin, zmax, '#',
                 10000, alpha, smpls, seed)
    psp = dict([(s,c) for s,c in psp.items() if c >= thpsp])
    n   = max([z for z,c in psp]+[0])
    border = list(range(n+1)); m = -1
    for z in reversed(border):  # determine the decision border
        m = max([c for a,c in psp if a == z]+[m])
        border[z] = m+1         # (support thresholds per size)
    border[0:2] = [oo,oo]       # entirely rule out sizes 0 and 1
    stderr.write('[%d signature(s)]' % len(psp))
    stderr.write(' done [%.2fs].\n' % (time()-t))

    # --- analyze (spike) trains with CoCoNAD ---
    if len(args) < 2: exit()    # check for an edge file name
    t = time()                  # start timer, print log message
    stderr.write('analyzing trains ... ')
    pats = coconad(trains, target, width, supp, zmin, zmax, 'a', '',
                   border=border)
    stderr.write('[%d pattern(s)]' % len(pats))
    stderr.write(' done [%.2fs].\n' % (time()-t))

    # --- write pattern output file ---
    t = time()                  # write the found patterns
    stderr.write('writing ' +args[1] +' ... ')
    with open(args[1], 'w') as out:
        for p,c in pats:        # traverse and print the patterns
            for n in p[:-1]: out.write(n +isep)
            out.write(p[-1])    # write the items of the pattern
            out.write(outfmt % c)
            out.write('\n')     # write the support information
    stderr.write('[%d pattern(s)]' % len(pats))
    stderr.write(' done [%.2fs].\n' % (time()-t))

    # --- build item sequence ---
    if len(args) < 3: exit()    # check for a sequence file name
    t = time()                  # start timer, print log message
    stderr.write('building item sequence ... ')
    iseq = itemseq(mode, parm, pats, trains, width)
    stderr.write('[%d item(s)]' % len(iseq))
    stderr.write(' done [%.2fs].\n' % (time()-t))

    # --- write item output file ---
    t = time()                  # start timer, print log message
    stderr.write('writing %s ... ' % args[2])
    with open(args[2], 'w') as out:
        for i,c in iseq:        # traverse the weigted items
            out.write(ishdr)    # print the record header
            out.write(str(i))   # write the item identifier
            out.write(outfmt % c)
            out.write('\n')     # write the support information
    stderr.write('[%d item(s)]' % len(iseq))
    stderr.write(' done [%.2fs].\n' % (time()-t))

    # --- cut best items at start ---
    #steps = [x[1]-y[1] for x,y in zip(iseq,iseq[1:]+[(0,0)])]
    #imod  = [(x[0],s*x[1]) for s,x in zip(steps,iseq)]
    #for x,y in zip(iseq,imod):
    #    print(x,y)
    
    best = max((iseq[i][1],i) for i in range(len(iseq)))[1] +1
    print sorted([int(i) for i,c in iseq[:best]]), ':', best

    #best = max((imod[i][1],i) for i in range(len(imod)))[1] +1
    #print sorted([int(i) for i,c in imod[:best]]), ':', best
